"""Allows CloudWanderer to store resources in DynamoDB."""
import datetime
import itertools
import json
import logging
import operator
import os
import pathlib
import sys
from functools import reduce
from random import randrange
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Iterable, Iterator, Optional, Union, cast

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

import boto3
from boto3.dynamodb.conditions import Attr, ConditionBase, Key

if TYPE_CHECKING:
    from mypy_boto3_dynamodb import DynamoDBServiceResource
else:
    DynamoDBServiceResource = object

from ..cloud_wanderer_resource import CloudWandererResource
from ..urn import URN
from ..utils import standardise_data_types
from .base_connector import ISO_DATE_FORMAT, BaseStorageConnector

logger = logging.getLogger(__name__)


class DynamoDBQueryArgs(TypedDict, total=False):
    """Valid DynamoDB Query args to facilitate type hinting."""

    Select: Union[
        Literal["ALL_ATTRIBUTES"], Literal["ALL_PROJECTED_ATTRIBUTES"], Literal["COUNT"], Literal["SPECIFIC_ATTRIBUTES"]
    ]
    KeyConditionExpression: Optional[Union[str, ConditionBase]]
    FilterExpression: Optional[Union[str, ConditionBase]]
    IndexName: Optional[str]


def _gen_resource_type_index(service: str, resource_type: str) -> str:
    """Generate a hash key for the resource type index.

    Arguments:
        service (str): The service name (e.g. ``'ec2'``)
        resource_type (str): The resource type (e.g. ``'instance'``)
    """
    return f"{service}#{resource_type}"


def _gen_resource_type_range(account_id: str, region: Optional[str]) -> str:
    """Generate a range key for the resource type index.

    Arguments:
        account_id (str): The AWS Account ID
        region (str): The AWS region (e.g. ``'eu-west-1'``

    """
    return f"{account_id}#{region or ''}"


def _gen_resource_type_condition_expression(
    hash_key: str, account_id: str = None, region: str = None
) -> Union[ConditionBase]:
    """Generate a condition expression for the resource type index.

    Will match ONLY on hash_key if neither ``account_id`` nor ``region`` are specified.
    If ``account_id`` is specified without region it will match all records matching ``account_id``.
    If ``account_id`` and region are specified it will match records matching both.
    If region is specified without ``account_id`` it will match nothing.

    Arguments:
        hash_key (str): The key to match for ``_resource_type_index``
        account_id (str): The AWS account id to search for.
        region (str): The AWS region to search for (e.g. ``'eu-west-1'``)
    """
    condition_expression = Key("_resource_type_index").eq(hash_key)
    if not account_id:
        return condition_expression
    range_key = _gen_resource_type_range(account_id=account_id, region=region)
    return condition_expression & Key("_resource_type_range").begins_with(range_key)


def _primary_key_from_urn(urn: URN) -> str:
    """Create a DynamoDB Primary Key from a resource's URN.

    Arguments:
        urn (URN): The URN to generate the primary key from.
    """
    return f"resource#{urn}"


def _urn_from_primary_key(pk: str) -> URN:
    """Create an URN from a resource's primary key.

    Arguments:
        pk (str): The primary key from which to parse the URN.
    """
    return URN.from_string(pk.split("#")[1])


def _dynamodb_items_to_resources(items: Iterable[dict], loader: Callable) -> Iterator[CloudWandererResource]:
    """Convert a resource and its attributes dynamodb records to a ResourceDict.

    Arguments:
        items (Iterable[dict]): The list of records retrieved from dynamodb.
        loader (Callable): The method which can be used to fulfil the :meth:`CloudWandererResource.load`

    """
    for _, group in itertools.groupby(items, lambda x: x["_id"]):
        grouped_items = list(group)
        base_resource = next(resource for resource in grouped_items if resource["_attr"] == "BaseResource")
        dependent_resource_urns = [URN.from_string(urn) for urn in base_resource.get("_dependent_resource_urns", [])]
        parent_urn: Optional[URN] = None
        if "_parent_urn" in base_resource:
            parent_urn = URN.from_string(base_resource["_parent_urn"])

        yield CloudWandererResource(
            urn=_urn_from_primary_key(base_resource["_id"]),
            dependent_resource_urns=dependent_resource_urns,
            resource_data=_strip_dynamodb_attrs(base_resource),
            parent_urn=parent_urn,
            discovery_time=datetime.datetime.strptime(base_resource["_discovery_time"], ISO_DATE_FORMAT),
            loader=loader,
        )


def _strip_dynamodb_attrs(raw_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Remove any underscore prefixed keys as these are attributes we use to identify the DynamoDB record.

    Arguments:
        raw_dict: The raw dictionary of the DynamoDB record that needs cleaning.
    """
    return {k: v for k, v in raw_dict.items() if not k.startswith("_")}


class DynamoDbConnector(BaseStorageConnector):
    """CloudWanderer Storage Connector for DynamoDB.

    Arguments:
        table_name (str):
            The name of the table to store resources in.
        endpoint_url (str):
            Optional override endpoint url for DynamoDB.
        boto3_session (boto3.session.Session):
            Optional boto3 session to use to interact with DynamoDB.
            Useful if your DynamoDB table is in a different account/region to your configured defaults.
        number_of_shards (int):
            The number of shards to break records across low-cardinality indices.
            Prevents hot-partitions. If you don't know what this means, ignore this setting.
        client_args (dict): Arguments to pass into the boto3 client.
            See: :meth:`boto3.session.Session.client`

    Example:
        >>> import cloudwanderer
        >>> cloud_wanderer = cloudwanderer.CloudWanderer(
        ...     storage_connectors=[cloudwanderer.storage_connectors.DynamoDbConnector(
        ...         endpoint_url='http://localhost:8000'
        ...     )]
        ... )
    """

    def __init__(
        self,
        table_name: str = "cloud_wanderer",
        endpoint_url: str = None,
        boto3_session: boto3.session.Session = None,
        client_args: dict = None,
        number_of_shards: int = 10,
    ) -> None:
        """Initialise the DynamoDbConnector.

        Arguments:
            table_name (str):
                The name of the table to read/write from/to.
            endpoint_url (str):
                Optional endpoint url (useful primarily to write to a local DynamoDB).
            boto3_session (str):
                Optional boto3 session to use to persist params.
            client_args (dict):
                Optional dictionary of arguments to be passed to the boto3 dynamodb client.
            number_of_shards (int):
                Optional specification of the number of shards to create for low-cardinality indexes.

        """
        self.client_args = client_args or {}
        if endpoint_url:
            self.client_args["endpoint_url"] = endpoint_url
        self.boto3_session = boto3_session or boto3.session.Session()
        self.table_name = table_name
        self.number_of_shards = number_of_shards
        self.dynamodb: DynamoDBServiceResource = self.boto3_session.resource("dynamodb", **self.client_args)
        self.dynamodb_table = self.dynamodb.Table(table_name)

    def init(self) -> None:
        """Create the DynamoDB Database."""
        table_creator = DynamoDbTableCreator(boto3_dynamodb_resource=self.dynamodb, table_name=self.table_name)
        table_creator.create_table()

    def write_resource(self, resource: CloudWandererResource) -> None:
        logger.debug(f"Writing: {resource.urn} to {self.table_name}")
        if resource.urn.is_partial:
            raise ValueError("Expected complete urn got partial for resource URN: %s.", resource.urn)
        item = {
            **self._generate_urn_index_values(cast(URN, resource.urn)),
            **standardise_data_types(resource.cloudwanderer_metadata.resource_data or {}),
            **{
                "_dependent_resource_urns": [str(urn) for urn in resource.dependent_resource_urns],
                "_discovery_time": resource.discovery_time.isoformat(),
            },
        }
        if resource.is_dependent_resource:
            item["_parent_urn"] = str(resource.parent_urn)
        self.dynamodb_table.put_item(Item=item)

    def _generate_urn_index_values(self, urn: URN, attr: str = "BaseResource") -> Dict[str, Any]:
        values = {
            "_id": _primary_key_from_urn(urn),
            "_attr": attr,
            "_urn": str(urn),
            "_resource_type": urn.resource_type,
            "_account_id": urn.account_id,
            "_region": urn.region,
            "_service": urn.service,
            "_resource_type_range": _gen_resource_type_range(urn.account_id, urn.region),
        }
        if attr == "BaseResource":
            values.update(
                {
                    "_resource_type_index": self._gen_shard(_gen_resource_type_index(urn.service, urn.resource_type)),
                    "_account_id_index": self._gen_shard(urn.account_id),
                }
            )
        return values

    def read_resource(self, urn: URN) -> Optional[CloudWandererResource]:
        """Return the resource with the specified :class:`cloudwanderer.urn.URN`.

        Arguments:
            urn (URN): The AWS URN of the resource to return
        """
        result = self._paginated_query(
            DynamoDBQueryArgs(KeyConditionExpression=Key("_id").eq(_primary_key_from_urn(urn)))
        )
        return next(_dynamodb_items_to_resources(result, loader=self.read_resource), None)

    def read_resources(
        self,
        cloud_name: str = None,
        account_id: str = None,
        region: str = None,
        service: str = None,
        resource_type: str = None,
        urn: URN = None,
    ) -> Iterator["CloudWandererResource"]:
        query_generator = DynamoDbQueryGenerator(cloud_name, account_id, region, service, resource_type, urn)
        for condition_expression in query_generator.condition_expressions:
            query_args = DynamoDBQueryArgs(
                KeyConditionExpression=condition_expression,
            )
            if query_generator.index is not None:
                query_args["IndexName"] = query_generator.index
                query_args["Select"] = "ALL_PROJECTED_ATTRIBUTES"
            if query_generator.condition_expressions is not None:
                query_args["FilterExpression"] = query_generator.filter_expression

            yield from _dynamodb_items_to_resources(self._paginated_query(query_args), loader=self.read_resource)

    def _paginated_query(self, query_args: DynamoDBQueryArgs) -> Generator[Dict[str, Any], None, None]:
        paginator = self.dynamodb.meta.client.get_paginator("query")
        pages = paginator.paginate(TableName=self.dynamodb_table.name, **query_args)  # type: ignore
        yield from (item for result in pages for item in result["Items"])

    def read_all(self) -> Iterator[dict]:
        """Return raw data from all DynamoDB table records (not just resources)."""
        paginator = self.dynamodb_table.meta.client.get_paginator("scan")
        yield from (item for page in paginator.paginate(TableName=self.dynamodb_table.name) for item in page["Items"])

    def delete_resource(self, urn: URN) -> None:
        """Delete the resource and all its resource attributes from DynamoDB.

        Arguments:
            urn (URN): The URN of the resource to delete from Dynamo
        """
        resource_records = itertools.chain(
            self._paginated_query(DynamoDBQueryArgs(KeyConditionExpression=Key("_id").eq(_primary_key_from_urn(urn)))),
            self._paginated_query(
                DynamoDBQueryArgs(IndexName="parent_urn", KeyConditionExpression=Key("_parent_urn").eq(str(urn)))
            ),
        )
        with self.dynamodb_table.batch_writer() as batch:
            for record in resource_records:
                logger.info("Deleting %s", record["_id"])
                batch.delete_item(Key={"_id": record["_id"], "_attr": record["_attr"]})

    def delete_resource_of_type_in_account_region(
        self,
        cloud_name: str,
        service: str,
        resource_type: str,
        account_id: str,
        region: str,
        cutoff: Optional[datetime.datetime],
    ) -> None:
        logger.debug("Deleting any %s discovered before %s", resource_type, cutoff)
        resource_records = self.read_resources(
            cloud_name=cloud_name, service=service, resource_type=resource_type, account_id=account_id, region=region
        )
        for resource in resource_records:
            if cutoff and resource.discovery_time >= cutoff:
                logger.debug("Skipping deletion of %s as it was discovered after our cutoff.", resource.urn)
                continue
            if resource.urn.is_partial:
                raise NotImplementedError(
                    "The DynamoDB Storage connector does not know how to delete partial URNs: %s.", resource.urn
                )
            logger.debug("Cleaning up %s discovered %s", str(resource.urn), resource.discovery_time)
            self.delete_resource(urn=cast(URN, resource.urn))

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    def _gen_shard(self, key: str, shard_id: int = None) -> str:
        """Append a shard designation to the end of a supplied key.

        Arguments:
            key (str): The key to which to append a shard id.
            shard_id (int): An optional shard id to append. If omitted a random shard will be generated.

        """
        shard_id = shard_id if shard_id is not None else randrange(self.number_of_shards - 1)
        return f"{key}#shard{shard_id}"

    def __repr__(self) -> str:
        """Return an instantiable string representation of this class."""
        return (
            f"{self.__class__.__name__}("
            f'table_name="{self.table_name}", '
            f'endpoint_url="{self.client_args.get("endpoint_url")}", '
            f'boto3_session="{self.boto3_session}", '
            f'client_args="{self.client_args}, '
            f"number_of_shards={self.number_of_shards}"
            ")"
        )

    def __str__(self) -> str:
        """Return a string representation of this class."""
        return f"<{self.__class__.__name__}={self.table_name}>"


class DynamoDbQueryGenerator:
    """Generate ConditionExpression and index name based on init params."""

    def __init__(
        self,
        cloud_name: str = None,
        account_id: str = None,
        region: str = None,
        service: str = None,
        resource_type: str = None,
        urn: URN = None,
        number_of_shards: int = 10,
    ) -> None:
        """Initialise QueryGenerator.

        Arguments:
            cloud_name: The name of the cloud
            account_id (str): AWS Account ID
            region (str): AWS region (e.g. ``'eu-west-2'``)
            service (str): Service name (e.g. ``'ec2'``)
            resource_type (str): Resource Type (e.g. ``'instance'``)
            urn (URN): Urn of the resource to retrieve
            number_of_shards (int): The number of shards we need to query in the table
        """
        self.cloud_name = cloud_name
        self.account_id = account_id
        self.region = region
        self.service = service
        self.resource_type = resource_type
        self.urn = urn
        self.number_of_shards = number_of_shards

    @property
    def index(self) -> Optional[str]:
        """Return the DynamoDB index to query.

        Raises:
            IndexNotAvailableException: Occurs when there is a no index that can fulfil this query.
                You will need to use a different set of arguments to query your resource if you receive this error.
        """
        index = None

        if self.service is not None and self.resource_type is not None:
            return "resource_type"
        if self.account_id is not None:
            return "account_id"
        if self.urn is not None:
            return index
        raise IndexNotAvailableException()

    @property
    def condition_expressions(self) -> Iterator[ConditionBase]:
        """Return the condition expression for the query.

        Raises:
            ValueError: If not all required arguments are passed for a given combination
        """
        if self.index is None and self.urn is not None:
            logger.info("Querying for %s", _primary_key_from_urn(self.urn))
            yield Key("_id").eq(_primary_key_from_urn(self.urn))
            return
        if self.index == "resource_type":
            if self.resource_type is None or self.service is None:
                raise ValueError("service, and resource_type must be specified when searching by resource type.")
            unsharded_key = _gen_resource_type_index(service=self.service, resource_type=self.resource_type)
            for sharded_key in self._yield_shards(unsharded_key):
                yield _gen_resource_type_condition_expression(
                    sharded_key, account_id=self.account_id, region=self.region
                )
            return
        if self.index == "account_id":
            if self.account_id is None:
                raise ValueError("account_id must be specified when searching by account id.")
            yield from [Key("_account_id_index").eq(shard) for shard in self._yield_shards(self.account_id)]
            return

    @property
    def filter_expression(self) -> ConditionBase:
        """Return a DynamoDB filter expression to use to filter out unwanted resources returned on our index."""
        query_args = ["account_id", "region", "service", "resource_type", "urn"]
        filter_elements = []
        for key in query_args:
            value = getattr(self, key)
            if value is not None:
                filter_elements.append(Attr(f"_{key}").eq(str(value)))
        return reduce(operator.and_, filter_elements)

    def _yield_shards(self, key: str) -> Generator[str, None, None]:
        for shard_id in range(0, self.number_of_shards):
            yield f"{key}#shard{shard_id}"


class DynamoDbTableCreator:
    """DynamoDB Table Creator class.

    Arguments:
        boto3_dynamodb_resource:
            The dynamodb resource object from boto3.
        table_name (str):
            The name of the table to create.

    """

    schema_file = os.path.join(pathlib.Path(__file__).parent.absolute(), "dynamodb_schema.json")

    def __init__(self, boto3_dynamodb_resource: DynamoDBServiceResource, table_name: str) -> None:
        """Initialise the DynamoDB Table Creator.

        Arguments:
            boto3_dynamodb_resource:
                The Boto3 DynamoDB resource object to use to create the table.
            table_name:
                The name we want the table to have.
        """
        self.dynamodb = boto3_dynamodb_resource
        self.dynamodb_table = self.dynamodb.Table(table_name)
        self.table_name = table_name
        self._schema = None

    def create_table(self) -> None:
        """Create the DynamoDB table."""
        logger.info(
            "Creating table in %s via %s",
            self.dynamodb.meta.client.meta.region_name,
            self.dynamodb.meta.client.meta.endpoint_url,
        )
        try:
            self.dynamodb.create_table(**{**self.schema["table"], **{"TableName": self.table_name}})
        except self.dynamodb_table.meta.client.exceptions.ResourceInUseException:
            self.dynamodb_table.load()
            logger.info("Table %s already exists, skipping creation.", self.table_name)

    @property
    def schema(self) -> dict:
        """Return the DynamoDB Schema.

        Raises:
            FileNotFoundError: When we cannot find the schema file.
        """
        if not self._schema:
            with open(self.schema_file) as schema_file:
                self._schema = json.load(schema_file)
        if self._schema is None:
            raise FileNotFoundError(f"DynamoDB Scheme not found at {self.schema_file}")
        return self._schema


class IndexNotAvailableException(Exception):
    """There is no DynamoDB index available for this type of query."""



"""Subclass of Boto3 Session class to provide additional helper methods."""
import logging
from typing import List

import boto3
from botocore.loaders import Loader

from ..cache_helpers import memoized_method
from .boto3_loaders import MergedServiceLoader
from .resource_factory import CloudWandererResourceFactory

logger = logging.getLogger(__name__)


class CloudWandererBoto3Session(boto3.session.Session):
    """Subclass of Boto3 Session class to provide additional helper methods."""

    def __init__(
        self,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        aws_session_token=None,
        region_name=None,
        botocore_session=None,
        profile_name=None,
        resource_factory=None,
        service_mapping_loader: Loader = None,
    ) -> None:
        self.service_mapping_loader = service_mapping_loader
        super().__init__(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            botocore_session=botocore_session,
            profile_name=profile_name,
        )

        self.resource_factory = resource_factory or CloudWandererResourceFactory(
            self._session.get_component("event_emitter"),
            service_mapping_loader=self._loader,
            cloudwanderer_boto3_session=self,
        )

    @memoized_method()
    def get_account_id(self) -> str:
        """Return the AWS Account ID our Boto3 session is authenticated against."""
        sts = self.client("sts")
        return sts.get_caller_identity()["Account"]

    def _setup_loader(self) -> None:
        """Create loader paths so that we can load resources."""
        self._loader = self.service_mapping_loader or MergedServiceLoader()

    @memoized_method()
    def get_enabled_regions(self) -> List[str]:
        """Return a list of enabled regions in this account."""
        regions = self.client("ec2").describe_regions()["Regions"]
        return [region["RegionName"] for region in regions if region["OptInStatus"] != "not-opted-in"]



"""Models specific to the CloudWanderer aws interface."""
from typing import Any, Dict, NamedTuple


class ResourceFilter(NamedTuple):
    """A definition for filtering a resource."""

    service_name: str
    resource_type: str
    filters: Dict[str, Any]



"""Main cloudwanderer module."""
import concurrent.futures
import logging
from datetime import datetime
from typing import Callable, Dict, List, NamedTuple, Union

from cloudwanderer.models import ServiceResourceType

from .aws_interface import CloudWandererAWSInterface
from .cloud_wanderer_resource import CloudWandererResource
from .storage_connectors import BaseStorageConnector
from .urn import URN, PartialUrn
from .utils import exception_logging_wrapper

logger = logging.getLogger("cloudwanderer")


class CloudWanderer:
    """CloudWanderer."""

    def __init__(
        self, storage_connectors: List["BaseStorageConnector"], cloud_interface: CloudWandererAWSInterface = None
    ) -> None:
        """Initialise CloudWanderer.

        Args:
            storage_connectors (List[BaseStorageConnector]):
                CloudWanderer storage connector objects.
            cloud_interface (CloudWandererAWSInterface):
                The cloud interface to get resources from.
                Defaults to :class:`~cloudwanderer.aws_interface.CloudWandererAWSInterface`.
        """
        self.storage_connectors = storage_connectors
        self.cloud_interface = cloud_interface or CloudWandererAWSInterface()

    def write_resource(self, urn: URN, **kwargs) -> None:
        """Fetch data for and persist to storage a single resource and its subresources.

        If the resource does not exist it will be deleted from the storage connectors.

        Arguments:
            urn (URN):
                The URN of the resource to write
            **kwargs:
                All additional keyword arguments will be passed down to the cloud interface client calls.
        """
        for storage_connector in self.storage_connectors:
            storage_connector.open()
        resources = list(self.cloud_interface.get_resource(urn=urn, **kwargs))

        for resource in resources:
            self._write_resource(resource=resource)
        if not resources:
            for storage_connector in self.storage_connectors:
                storage_connector.delete_resource(urn)

        for storage_connector in self.storage_connectors:
            storage_connector.close()

    def write_resources(
        self,
        regions: List[str] = None,
        service_resource_types: List[ServiceResourceType] = None,
        **kwargs,
    ) -> None:
        """Write all AWS resources in this account from all regions and all services to storage.

        All arguments are optional.

        Arguments:
            regions:
                The name of the region to get resources from (defaults to session default if not specified)
            service_resource_types:
                The resource types to discover.
            kwargs:
                All additional keyword arguments will be passed down to the cloud interface client calls.

        Raises:
            ValueError: If invalid get/delete urns are produced by the cloud interface's get_resource_discovery_actions
        """
        for storage_connector in self.storage_connectors:
            storage_connector.open()
        action_sets = self.cloud_interface.get_resource_discovery_actions(
            regions=regions, service_resource_types=service_resource_types
        )
        discovery_start_times: Dict[str, datetime] = {}
        for action_set in action_sets:
            for get_urn in action_set.get_urns:
                if not get_urn.region or not get_urn.service or not get_urn.resource_type:
                    raise ValueError(f"Invalid get_urn {get_urn}")
                resources = self.cloud_interface.get_resources(
                    region=get_urn.region,
                    service_name=get_urn.service,
                    resource_type=get_urn.resource_type,
                )
                for resource in resources:
                    earliest_resource_discovered = discovery_start_times.get(resource.urn.cloud_service_resource_label)
                    if not earliest_resource_discovered or resource.discovery_time < earliest_resource_discovered:
                        discovery_start_times[resource.urn.cloud_service_resource_label] = resource.discovery_time
                    self._write_resource(resource)
            for delete_urn in action_set.delete_urns:
                if (
                    not delete_urn.account_id
                    or not delete_urn.region
                    or not delete_urn.service
                    or not delete_urn.resource_type
                    or not delete_urn.cloud_name
                ):
                    raise ValueError(f"Invalid delete_urn {delete_urn}")
                for storage_connector in self.storage_connectors:
                    storage_connector.delete_resource_of_type_in_account_region(
                        cloud_name=delete_urn.cloud_name,
                        account_id=delete_urn.account_id,
                        region=delete_urn.region,
                        service=delete_urn.service,
                        resource_type=delete_urn.resource_type,
                        cutoff=discovery_start_times.get(delete_urn.cloud_service_resource_label),
                    )
        for storage_connector in self.storage_connectors:
            storage_connector.close()

    def write_resources_concurrently(
        self,
        cloud_interface_generator: Callable,
        storage_connector_generator: Callable,
        concurrency: int = 10,
        **kwargs,
    ) -> List["CloudWandererConcurrentWriteThreadResult"]:
        """Write all AWS resources in this account from all regions and all services to storage.

        Any additional args will be passed into the cloud interface's ``get_`` methods.
        **WARNING:** Experimental.

        Arguments:
            concurrency (int):
                Number of query threads to invoke concurrently.
            cloud_interface_generator (Callable):
                 A method which returns a new cloud interface session when called.
                This helps prevent non-threadsafe cloud interfaces from interfering with each others.
            storage_connector_generator (Callable):
                A method which returns a list of storage connectors when called.
                The returned connectors should be instances of the same connectors each time the method is called.
                These connectors do **not** need to be thread safe and will be returned at the end of execution.
            **kwargs:
                Additional keyword arguments will be passed down to the cloud interface methods.
        """
        logger.info("Writing resources in all regions")
        logger.warning("Using concurrency of: %s - CONCURRENCY IS EXPERIMENTAL", concurrency)
        with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
            threads = []
            for region_name in self.cloud_interface.get_enabled_regions():
                cw = CloudWanderer(
                    storage_connectors=storage_connector_generator(), cloud_interface=cloud_interface_generator()
                )
                threads.append(
                    executor.submit(
                        exception_logging_wrapper,
                        method=cw.write_resources,
                        regions=[region_name],
                        return_value=cw.storage_connectors,
                        **kwargs,
                    )
                )
        thread_results: List[CloudWandererConcurrentWriteThreadResult] = []
        for thread in threads:
            result = thread.result()
            if result:
                thread_results.append(CloudWandererConcurrentWriteThreadResult(storage_connectors=result))
        return thread_results

    def _write_resource(self, resource: CloudWandererResource) -> Union[URN, PartialUrn]:
        for storage_connector in self.storage_connectors:
            storage_connector.write_resource(resource)
        return resource.urn


class CloudWandererConcurrentWriteThreadResult(NamedTuple):
    """The result from write_resources_concurrently."""

    storage_connectors: BaseStorageConnector



"""Create the CloudWandererServiceResource objects that do the magic."""
import logging
import re
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Optional, Type

import jmespath  # type: ignore
from boto3.resources.base import ServiceResource
from boto3.resources.collection import CollectionManager
from boto3.resources.factory import ResourceFactory
from boto3.resources.model import Collection
from boto3.resources.params import create_request_parameters
from botocore import xform_name
from botocore.loaders import Loader
from botocore.model import Shape

from ..exceptions import UnsupportedResourceTypeError
from ..models import Relationship, RelationshipAccountIdSource, RelationshipRegionSource, TemplateActionSet
from ..urn import URN, PartialUrn
from ..utils import snake_to_pascal
from .boto3_helpers import _clean_boto3_metadata
from .boto3_loaders import MergedServiceLoader, ServiceMap

if TYPE_CHECKING:
    from .session import CloudWandererBoto3Session
    from .stubs.resource import CloudWandererServiceResource
    from .stubs.service_context import ServiceContext


logger = logging.getLogger(__name__)


class CloudWandererResourceFactory(ResourceFactory):
    """Enriches functionality of boto3 resource objects with CloudWanderer specific methods."""

    def __init__(
        self,
        emitter,
        cloudwanderer_boto3_session: "CloudWandererBoto3Session",
        service_mapping_loader: Optional[Loader] = None,
    ) -> None:
        super().__init__(emitter=emitter)
        self.service_mapping_loader = service_mapping_loader or MergedServiceLoader()
        self.cloudwanderer_boto3_session = cloudwanderer_boto3_session

    def load_from_definition(self, resource_name, single_resource_json_definition, service_context) -> type:
        class_definition = super().load_from_definition(resource_name, single_resource_json_definition, service_context)
        attrs: Dict[str, Any] = {}
        # CloudWanderer resource methods
        self._load_cloudwanderer_methods(
            attrs=attrs,
            resource_name=resource_name,
            service_context=service_context,
            original_class_definition=class_definition,
        )

        # CloudWanderer resource properties
        self._load_cloudwanderer_properties(
            attrs=attrs,
            resource_name=resource_name,
            service_context=service_context,
        )

        for attribute_name, attribute_value in attrs.items():
            setattr(class_definition, attribute_name, attribute_value)
        return class_definition

    def _load_cloudwanderer_methods(
        self,
        attrs: Dict[str, Any],
        resource_name: str,
        service_context: "ServiceContext",
        original_class_definition: Type[ServiceResource],
    ) -> None:
        if service_context.service_name != xform_name(resource_name):
            # This should only exist only exist on Resources, not on Services
            attrs["get_discovery_action_templates"] = self._create_get_discovery_action_templates()
            # attrs["get_dependent_resource"] = self._create_get_dependent_resource(service_context)
            attrs["get_secondary_attributes_map"] = self._create_get_secondary_attributes_map()
            if hasattr(original_class_definition, "load"):
                attrs["load"] = self._create_load(original_class_definition=original_class_definition)
        else:
            attrs["resource"] = self._create_resource()
        attrs["get_collection_manager"] = self._create_get_collection_manager()
        attrs["get_collection_model"] = self._create_get_collection_model()
        attrs["collection"] = self._create_collection_getter()
        attrs["get_account_id"] = self._create_get_account_id()
        attrs["get_urn"] = self._create_get_urn()
        attrs["get_region"] = self._create_get_region()
        attrs["get_secondary_attributes"] = self._create_get_secondary_attributes()

    def _create_load(self, original_class_definition: Any) -> Callable:
        if not hasattr(original_class_definition, "load"):
            raise ValueError("Cannot create load method for class which does not need one.")
        original_class_load = original_class_definition.load

        def load(self, *args) -> None:
            if self.meta.resource_model.load:
                identifiers = create_request_parameters(self, self.meta.resource_model.load.request)
            elif self.service_name in "s3" and self.resource_type in "bucket":
                # This is hardcoded in boto3 believe it or not!
                # https://github.com/boto/boto3/blob/master/boto3/s3/inject.py#L57
                identifiers = {"BucketName": self.name}
            else:
                raise UnsupportedResourceTypeError(
                    f"{self.service_name} {self.resource_type} does not have a load definition in its resources-1.json"
                )

            has_non_empty_values = any(list([x for x in identifiers.values()]))
            if not has_non_empty_values:
                logger.debug(
                    "Load is a noop on this %s %s because we are an empty_resource=True resource",
                    self.service_name,
                    self.resource_type,
                )
                return
            original_class_load(self)

        return load

    def _create_get_discovery_action_templates(self) -> Callable:
        def get_discovery_action_templates(self, discovery_regions: List[str]) -> List[TemplateActionSet]:
            """Return the discovery actions to be performed if getting this resource type in this region.

            Arguments:
                discovery_regions: The regions in which the query would be performed.
            """
            template_actions = []
            for discovery_region in discovery_regions:
                actions = TemplateActionSet([], [])
                if not self.resource_map.should_query_resources_in_region(discovery_region):
                    continue
                if self.service_map.is_global_service and self.resource_map.regional_resource:
                    cleanup_region = "ALL_REGIONS"
                else:
                    cleanup_region = discovery_region
                if self.resource_map.type != "dependentResource":
                    actions.get_urns.append(
                        PartialUrn(
                            cloud_name="aws",
                            service=self.service_map.name,
                            region=discovery_region,
                            resource_type=self.resource_type,
                        )
                    )
                actions.delete_urns.append(
                    PartialUrn(
                        cloud_name="aws",
                        service=self.service_map.name,
                        region=cleanup_region,
                        resource_type=self.resource_type,
                    )
                )
                template_actions.append(actions)
            return template_actions

        return get_discovery_action_templates

    def _create_get_collection_manager(self) -> Callable:
        def get_collection_manager(self, resource_type=str) -> CollectionManager:
            collection_model = self.get_collection_model(resource_type=resource_type)
            return getattr(self, collection_model.name)

        return get_collection_manager

    def _create_get_collection_model(self) -> Callable:
        def get_collection_model(self, resource_type) -> Optional[str]:
            """Resource names *almost* always match their resource type, but not always.

            I'm looking at you EC2 KeyPair!

            A resource _name_ is the PascalCase key it has in the ``resources`` dict in the service definition json.
            A resource _type_ is (in the argument to this method) a snake_case resource type which matches
                the resource name.
            A resource _type_ in boto3 is a PascalCase resource type used to associate a Collection (hasMany) with its
                resource definition.
            In order to identify a collection we need to:
                1. Get the boto3 resource type of the resource which has the resource name that corresponse with the
                    cloudwanderer resource type
                2. Lookup the collection that references that resource type.
                3. Return that collection name

            Arguments;
                resource_type: The snake case resource type to get a model for (e.g. ``'bucket'``)

            Raises:
                UnsupportedResourceTypeError: Tried to get a collection model for a resource which does not have one.
            """
            for collection_model in self.meta.resource_model.collections:
                if xform_name(collection_model.resource.type) == resource_type:
                    return collection_model
            raise UnsupportedResourceTypeError(f"Could not find Boto3 collection for {resource_type}")

        return get_collection_model

    def _create_resource(self) -> Callable:
        def resource(
            self, resource_type: str, identifiers: List[str] = None, empty_resource=False
        ) -> "CloudWandererServiceResource":
            """Get a Boto3 ServiceResource object for a resource that exists in this service.

            Specifying empty_resource=True will return a ServiceResource object which does not
            correspond to a specific resource in AWS but allows access to resource type metadata.

            Raises:
                UnsupportedResourceTypeError: Tried to get a collection model for a resource which does not have one.
            """
            for resource in self.meta.resource_model.subresources:
                resource_name = xform_name(resource.name)
                if resource_name == resource_type:
                    if empty_resource:
                        identifiers = ["" for _ in resource.resource.model.identifiers]
                    return getattr(self, resource.name)(*identifiers)
            raise UnsupportedResourceTypeError(f"Could not find Boto3 resource for {resource_type}")

        return resource

    def _create_collection_getter(self) -> Callable:
        def collection(self, resource_type: str, filters: Optional[Dict[str, str]] = None) -> Collection:
            filters = filters or {}
            try:
                collection_model = self.get_collection_model(resource_type)
            except UnsupportedResourceTypeError:
                # If it's not a collection it might be a 'reference'
                for reference in self.meta.resource_model.references:
                    if xform_name(reference.resource.type) == resource_type:
                        return getattr(self, reference.name)
                raise
            collection_model.name
            collection_manager = getattr(self, collection_model.name)
            return collection_manager.filter(**filters)

        return collection

    def _create_get_urn(self) -> Callable:
        def get_urn(self) -> URN:
            id_parts = [getattr(self, identifier) for identifier in self.meta.identifiers]
            return URN(
                account_id=self.get_account_id(),
                region=self.get_region(),
                service=self.service_name,
                resource_type=self.resource_type,
                resource_id_parts=id_parts,
            )

        return get_urn

    def _create_get_secondary_attributes_map(self) -> Callable[..., Dict[str, Any]]:
        def get_secondary_attributes_map(self) -> Dict[str, Any]:
            """Return a dictionary representation of this resource's secondary attributes."""
            result = {}
            for secondary_attribute in self.get_secondary_attributes():
                logger.debug("Getting secondary attribute: %s", secondary_attribute)
                for attribute_map in secondary_attribute.resource_map.secondary_attribute_maps:
                    result[attribute_map.destination_name] = jmespath.search(
                        attribute_map.source_path, secondary_attribute.normalized_raw_data
                    )
            return result

        return get_secondary_attributes_map

    def _create_get_secondary_attributes(self) -> Callable:
        def get_secondary_attributes(self) -> Generator["CloudWandererServiceResource", None, None]:
            for secondary_attribute_name in self.secondary_attribute_names:
                getter = getattr(self, snake_to_pascal(secondary_attribute_name))
                secondary_attribute_resource = getter()
                secondary_attribute_resource.load()
                yield secondary_attribute_resource

        return get_secondary_attributes

    def _create_secondary_attribute_names(self) -> property:
        def secondary_attribute_names(self) -> List[str]:
            secondary_attribute_names = []
            for subresource in self.meta.resource_model.subresources:
                resource_map = self.service_map.get_resource_map(xform_name(subresource.name))
                if not resource_map or resource_map.type != "secondaryAttribute":
                    continue
                secondary_attribute_names.append(xform_name(subresource.name))

            return secondary_attribute_names

        return property(secondary_attribute_names)

    def _create_get_account_id(self) -> Callable:
        def get_account_id(self) -> str:
            return self.cloudwanderer_boto3_session.get_account_id()

        return get_account_id

    def _create_get_region(self) -> Callable:
        def get_region(self) -> str:
            if self.resource_map.region_request:
                method = getattr(self.meta.client, self.resource_map.region_request.operation)
                result = method(**self.resource_map.region_request.build_params(self))
                return (
                    jmespath.search(self.resource_map.region_request.path_to_region, result)
                    or self.resource_map.region_request.default_value
                )
            if self.service_map.global_service:
                return self.service_map.global_service_region

            return self.meta.client.meta.region_name

        return get_region

    def _create_normalized_raw_data(self) -> property:
        def normalized_raw_data(self) -> Dict[str, Any]:
            """Return the raw data dictionary for this resource, ensuring that all possible keys are present."""
            result = {attribute: None for attribute in self.shape.members.keys()}
            result.update(self.meta.data or {})
            return _clean_boto3_metadata(result)

        return property(normalized_raw_data)

    def _create_resource_types(self) -> property:
        def resource_types(self) -> List[str]:
            """List resources which are directly enumerable from the service."""
            resource_types = [
                xform_name(collection.resource.type) for collection in self.meta.resource_model.collections
            ]

            return resource_types

        return property(resource_types)

    def _create_dependent_resource_types(self) -> property:
        def dependent_resource_types(self) -> List[str]:
            dependent_resource_types = set()
            for collection in self.meta.resource_model.collections:
                resource_type = xform_name(collection.resource.type)
                resource_map = self.service_map.get_resource_map(resource_type=resource_type)
                if resource_map.type == "dependentResource":
                    dependent_resource_types.add(resource_type)
            for subresource in self.meta.resource_model.subresources + self.meta.resource_model.references:
                resource_type = xform_name(subresource.resource.type)
                if resource_type in dependent_resource_types:
                    continue
                resource_map = self.service_map.get_resource_map(resource_type=resource_type)
                if resource_map.type == "dependentResource":
                    dependent_resource_types.add(resource_type)

            return list(dependent_resource_types)

        return property(dependent_resource_types)

    def _create_shape(self) -> property:
        def shape(self) -> Shape:
            service_model = self.meta.client.meta.service_model
            return service_model.shape_for(self.meta.resource_model.shape)

        return property(shape)

    def _create_relationships(self) -> property:
        def relationships(self) -> List[Relationship]:
            """Return PartialURNs for the relationships this resource has with other resources."""
            relationships = []
            for relationship_specification in self.resource_map.relationships:
                base_paths_raw = jmespath.search(
                    relationship_specification.base_path,
                    {
                        **self.normalized_raw_data,
                        **self.get_secondary_attributes_map(),
                    },
                )
                base_paths = [base_paths_raw] if not isinstance(base_paths_raw, list) else base_paths_raw
                for base_path in base_paths:
                    if not base_path:
                        logger.debug("Skipping building a relationship for %s as the basePath is empty", self.get_urn())
                        continue
                    urn_args = {
                        "cloud_name": "aws",
                        "account_id": "unknown",
                        "region": "unknown",
                        "service": relationship_specification.service,
                        "resource_type": relationship_specification.resource_type,
                        "resource_id_parts": [],
                    }

                    if relationship_specification.account_id_source == RelationshipAccountIdSource.SAME_AS_RESOURCE:
                        urn_args["account_id"] = self.get_account_id()

                    if relationship_specification.region_source == RelationshipRegionSource.SAME_AS_RESOURCE:
                        urn_args["region"] = self.get_region()

                    for id_part in relationship_specification.id_parts:
                        id_raw = jmespath.search(id_part.path, base_path)
                        if not id_raw:
                            continue
                        if not id_part.regex_pattern:
                            urn_args["resource_id_parts"].append(id_raw)
                            continue
                        result = re.match(id_part.regex_pattern, id_raw)
                        if not result:
                            continue
                        for arg_name, arg_value in result.groupdict().items():
                            if arg_name in ["cloud_name", "account_id", "region", "service", "resource_type"]:
                                urn_args[arg_name] = arg_value
                                continue
                            if arg_name.startswith("id_part_"):
                                urn_args["resource_id_parts"].append(arg_value)
                    if not urn_args["resource_id_parts"]:
                        continue
                    relationships.append(
                        Relationship(partial_urn=PartialUrn(**urn_args), direction=relationship_specification.direction)
                    )
            return relationships

        return property(relationships)

    def _create_is_dependent_resource(self) -> property:
        def is_dependent_resource(self) -> bool:
            return self.resource_map.type == "dependentResource"

        return property(is_dependent_resource)

    def _load_cloudwanderer_properties(
        self, attrs: Dict[str, Any], resource_name: str, service_context: "ServiceContext"
    ) -> None:
        attrs["service_name"] = service_context.service_name
        attrs["service_map"] = ServiceMap.factory(
            name=service_context.service_name,
            definition=self.service_mapping_loader.load_service_model(
                service_name=service_context.service_name, type_name="resources-cw-1", api_version=None
            ),
        )
        attrs["cloudwanderer_boto3_session"] = self.cloudwanderer_boto3_session

        if resource_name == service_context.service_name:  # type: ignore
            # If it is a service:
            attrs["resource_types"] = self._create_resource_types()
        else:
            # If it is a resource:
            attrs["normalized_raw_data"] = self._create_normalized_raw_data()
            attrs["resource_type"] = xform_name(resource_name)
            attrs["resource_map"] = attrs["service_map"].get_resource_map(resource_type=xform_name(resource_name))
            attrs["dependent_resource_types"] = self._create_dependent_resource_types()
            attrs["secondary_attribute_names"] = self._create_secondary_attribute_names()
            attrs["shape"] = self._create_shape()
            attrs["relationships"] = self._create_relationships()
            attrs["is_dependent_resource"] = self._create_is_dependent_resource()



"""Standardised dataclasses for returning resources from storage."""

import datetime
import logging
from typing import Any, Callable, Generator, List, Optional, Tuple, Union

from botocore import xform_name

from .models import Relationship  # type: ignore
from .urn import URN, PartialUrn

logger = logging.getLogger(__name__)


class ResourceMetadata:
    """Metadata for a :class:`CloudWandererResource`.

    Contains the original dictionaries of the resource and its attributes.

    Attributes:
        resource_data (dict):
            The raw dictionary representation of the Resource.
    """

    def __init__(self, resource_data: dict) -> None:
        """Initialise the data class.

        Arguments:
            resource_data (dict):
                The raw dictionary representation of the Resource.
        """
        self.resource_data = resource_data

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        """Allow this object to be converted to a dictionary."""
        yield from self.resource_data.items()


class CloudWandererResource:
    """A simple representation of a resource that prevents any storage metadata polluting the resource dictionary.

    This can be a resource, or a dependent resource.
    A dependent resource in CloudWanderer is a resource which does not have a unique identifier of its own and depends
    upon its parent for its identity.

    Attributes:
        urn: The URN of the resource.
        dependent_resource_urns: The URNs of this resource's dependent resources (e.g. role_policies for a role).
        relationships: The PartialURNs of resources that are related to this one.
        parent_urn: The URN of this resource's parent (only exists if this is a dependent resource).
        cloudwanderer_metadata (ResourceMetadata): The metadata of this resource (including attributes).
    """

    def __init__(
        self,
        urn: Union[URN, PartialUrn],
        resource_data: dict,
        relationships: List[Relationship] = None,
        loader: Optional[Callable] = None,
        dependent_resource_urns: List[URN] = None,
        parent_urn: URN = None,
        discovery_time: datetime.datetime = None,
    ) -> None:
        """Initialise the resource.

        Arguments:
            urn: The URN of the resource.
            resource_data: The dictionary containing the raw data about this resource.
            relationships: The relationships this resource has with other resources.
            loader: The method which can be used to fulfil the :meth:`CloudWandererResource.load`.
            dependent_resource_urns: The URNs of the dependent resources of this resource.
            parent_urn: The URN of the resource's parent resource.
            discovery_time: The time the resource was discovered.
        """
        self.urn = urn
        self.relationships = relationships or []
        self.dependent_resource_urns = dependent_resource_urns or []
        self.parent_urn = parent_urn
        self.cloudwanderer_metadata = ResourceMetadata(resource_data=resource_data or {})
        self.discovery_time = discovery_time or datetime.datetime.now()

        self._loader = loader
        self._set_resource_data_attrs()

    def load(self) -> None:
        """Inflate this resource with all data from the original storage connector it was spawned from.

        Raises:
            ValueError: Occurs if the storage connector loader isn't populated or the resource no longer exists
                in the StorageConnector's storage.
        """
        if self._loader is None:
            raise ValueError(f"Could not inflate {self}, storage connector loader not populated")
        updated_resource = self._loader(urn=self.urn)
        if updated_resource is None:
            raise ValueError(f"Could not inflate {self}, does not exist in storage")
            return
        self.cloudwanderer_metadata = updated_resource.cloudwanderer_metadata
        self._set_resource_data_attrs()

    @property
    def is_inflated(self) -> bool:
        """Return whether this resource has all the attributes from storage."""
        return bool([key for key in self.cloudwanderer_metadata.resource_data if not key.startswith("_")])

    @property
    def is_dependent_resource(self) -> bool:
        return bool(self.parent_urn)

    def _set_resource_data_attrs(self) -> None:
        for key, value in self.cloudwanderer_metadata.resource_data.items():
            if key.startswith("_"):
                continue
            setattr(self, xform_name(key), value)

    def __repr__(self) -> str:
        """Return a code representation of this resource."""
        return str(
            f"{self.__class__.__name__}("
            f"urn={repr(self.urn)}, "
            f"dependent_resource_urns={repr(self.dependent_resource_urns)}, "
            f"resource_data={self.cloudwanderer_metadata.resource_data})"
        )

    def __str__(self) -> str:
        """Return the string representation of this Resource."""
        return repr(self)

    def __eq__(self, other) -> bool:
        """Return the equality or not of the compared object.

        Arguments:
            other: The resource to compare our equality against.
        """
        return repr(self) == repr(other)

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        """Allow this object to be converted to a dictionary."""
        for attribute_name in vars(self):
            if attribute_name.startswith("_"):
                continue
            value = getattr(self, attribute_name)
            if isinstance(value, ResourceMetadata):
                value = dict(value)
            yield attribute_name, value



"""Helper classes and methods for interacting with boto3."""


def _clean_boto3_metadata(boto3_metadata: dict) -> dict:
    """Remove unwanted keys from boto3 metadata dictionaries.

    Arguments:
        boto3_metadata (dict): The raw dictionary of metadata typically found in resource.meta.data
    """
    boto3_metadata = boto3_metadata or {}
    unwanted_keys = ["ResponseMetadata"]
    for key in unwanted_keys:
        if key in boto3_metadata:
            del boto3_metadata[key]
    return boto3_metadata



"""A dataclass for building and querying AWS URNs.

AWS URNs are a standardised string format (name borrowed from Pulumi) which provide all the information required to
find a resource in AWS, whereas AWS ARNs do not always provide this information.

.. code-block ::

    # Format
    'urn:aws:<account_id>:<region>:<service>:<resource_type>:<resource_id>'
    # e.g.
    'urn:aws:111111111111:eu-west-2:iam:vpc:vpc-11111111'

Example:
    >>> from cloudwanderer import URN
    >>> URN.from_string('urn:aws:111111111111:eu-west-2:iam:vpc:vpc-11111111')
    URN(cloud_name='aws', account_id='111111111111', region='eu-west-2', \
service='iam', resource_type='vpc', resource_id_parts=['vpc-11111111'])

"""
import re
from typing import Any, Generator, List, Optional, Tuple


class PartialUrn:
    """A partially specified URN.

    Useful for matching unknown or multiple URNs.
    """

    cloud_name: Optional[str]
    account_id: Optional[str]
    region: Optional[str]
    service: Optional[str]
    resource_type: Optional[str]
    resource_id_parts: List[str]

    def __init__(
        self,
        cloud_name: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
        service: Optional[str] = None,
        resource_type: Optional[str] = None,
        resource_id_parts: Optional[List[str]] = None,
    ) -> None:
        self.cloud_name = cloud_name
        self.account_id = account_id
        self.region = region
        self.service = service
        self.resource_type = resource_type
        self.resource_id_parts: List[str] = resource_id_parts or []
        self.resource_id = "/".join(self.escape_id(id_part) or "" for id_part in self.resource_id_parts)

    def copy(
        self,
        cloud_name: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
        service: Optional[str] = None,
        resource_type: Optional[str] = None,
        resource_id_parts: Optional[List[str]] = None,
    ) -> "PartialUrn":
        return PartialUrn(
            cloud_name=cloud_name or self.cloud_name,
            account_id=account_id or self.account_id,
            region=region or self.region,
            service=service or self.service,
            resource_type=resource_type or self.resource_type,
            resource_id_parts=resource_id_parts or self.resource_id_parts,
        )

    @property
    def is_partial(self) -> bool:
        for value in self.__dict__.values():
            if value == "unknown":
                return True
        return False

    @property
    def is_dependent_resource(self) -> bool:
        if not self.resource_id_parts:
            raise ValueError(
                "Cannot determine whether this PartialURN is dependent because it has no resource id parts"
            )
        return len(self.resource_id_parts) > 1

    @property
    def cloud_service_resource_label(self) -> str:
        """Return the cloud service resource label (e.g. ``aws_iam_role``).

        Raises:
            ValueError: If we do not have sufficient information to generate a label
        """
        if (
            not isinstance(self.cloud_name, str)
            or not isinstance(self.service, str)
            or not isinstance(self.resource_type, str)
        ):
            raise ValueError(
                "Cannot determine this PartialURN's "
                "cloud_service_resource as it is missing one of the required attributes"
            )
        return "_".join([self.cloud_name, self.service, self.resource_type]).lower()

    @staticmethod
    def unescape_id(escaped_id: Optional[str]) -> Optional[str]:
        """Return an unescaped ID with the forward slashes unescaped.

        Arguments:
            escaped_id: The id to unescape.
        """
        if escaped_id is None:
            return None
        return re.sub(r"\\(/|:)", r"\1", escaped_id)

    @staticmethod
    def escape_id(unescaped_id: Optional[str]) -> Optional[str]:
        """Return an escaped ID with the forward slashes escaped.

        Arguments:
            unescaped_id: The id to escape.
        """
        if unescaped_id is None:
            return None
        if not isinstance(unescaped_id, str):
            unescaped_id = str(unescaped_id)
        return re.sub(r"(?<!\\)(/|:)", r"\\\1", unescaped_id)

    def __str__(self) -> str:
        """Return a string representation of the URN."""
        base = ":".join(
            [
                str(part or "unknown")
                for part in ["urn", self.cloud_name, self.account_id, self.region, self.service, self.resource_type]
            ]
        )
        return f"{base}:{self.resource_id}"

    def __repr__(self) -> str:
        """Return a class representation of the URN."""
        return str(
            f"{self.__class__.__name__}("
            f"cloud_name='{ self.cloud_name}', "
            f"account_id='{self.account_id}', "
            f"region='{self.region}', "
            f"service='{self.service}', "
            f"resource_type='{self.resource_type}', "
            f"resource_id_parts={self.resource_id_parts})"
        )

    def __eq__(self, other: Any) -> bool:
        """Allow comparison of one URN to another.

        Arguments:
            other (Any): The other object to compare this one with.
        """
        return str(self) == str(other)

    def __iter__(self) -> Generator[Tuple[str, str], None, None]:
        """Allow the URN to be turned into a dict."""
        for attribute_name in vars(self):
            if attribute_name.startswith("_"):
                continue
            yield attribute_name, getattr(self, attribute_name)


class URN(PartialUrn):
    """A dataclass for building and querying AWS URNs."""

    account_id: str
    region: str
    service: str
    resource_type: str
    resource_id: str

    def __init__(
        self,
        account_id: str,
        region: str,
        service: str,
        resource_type: str,
        resource_id_parts: Optional[list] = None,
        cloud_name: str = None,
    ) -> None:
        """Initialise an AWS Urn.

        Arguments:
            account_id: AWS Account ID (e.g. ``111111111111``).
            region: AWS region (e.g. `us-east-1``).
            service: AWS Service (e.g. ``iam``).
            resource_type: AWS Resource Type (e.g. ``role_policy``)
            resource_id_parts: AWS Resource Id (e.g. ``test-role-policy``)
            cloud_name: The name of the cloud this resource exists in (defaults to ``'aws'``)

        Attributes:
            account_id: AWS Account ID (e.g. ``111111111111``).
            region: AWS region (e.g. `us-east-1``).
            service: AWS Service (e.g. ``iam``).
            resource_type: AWS Resource Type (e.g. ``role_policy``)
            resource_id: AWS Resource Id (e.g. ``test-role-policy``)
            resource_id_parts: AWS Resource ID parts (e.g. ``['test-role', 'test-role-policy']``)
            cloud_name: The name of the cloud this resource exists in (defaults to ``'aws'``)

        Raises:
            ValueError: If incorrect values are supplied.
        """
        if not resource_id_parts or not all(resource_id_parts):
            raise ValueError("resource_id or id_parts must be supplied with non empty values")

        super().__init__(
            cloud_name=cloud_name or "aws",
            account_id=account_id,
            region=region,
            service=service,
            resource_type=resource_type,
            resource_id_parts=resource_id_parts,
        )

    @classmethod
    def from_string(cls, urn_string: str) -> "URN":
        """Create an URN Object from an URN string.

        Arguments:
            urn_string (str): The string version of an AWSUrn to convert into an object.

        Returns:
            URN: The instantiated AWS URN.

        Raises:
            ValueError: When no valid resource id found
        """
        parts = re.split(r"(?<!\\):", urn_string)
        try:
            resource_id_parts = re.split(r"(?<!\\)/", parts[6])
        except IndexError:
            raise ValueError("Resource ID must be supplied as the 7th element in a colon separated string")
        return cls(
            account_id=parts[2],
            region=parts[3],
            service=parts[4],
            resource_type=parts[5],
            resource_id_parts=[cls.unescape_id(id_part) for id_part in resource_id_parts],
        )



"""Models for CloudWanderer data."""
from enum import Enum, auto, unique
from typing import List, NamedTuple

from .urn import PartialUrn


class ActionSet(NamedTuple):
    """Define a list of partial URNs to discover and delete."""

    get_urns: List[PartialUrn]
    delete_urns: List[PartialUrn]


class TemplateActionSet(ActionSet):
    """An set of actions which have not yet had all their placeholders inflated.

    This differs from a regular ActionSet action insofar as it
    will probably contain actions with the region 'ALL'.
    These actions need to be unpacked into region specific actions that
    reflect the enabled regions in the AWS account in question
    before being placed into a non-cloud specific ActionSet class
    for CloudWanderer to consume.
    """

    def inflate(self, regions: List[str], account_id: str) -> ActionSet:
        new_action_set = ActionSet(get_urns=[], delete_urns=[])
        for partial_urn in self.get_urns:
            new_action_set.get_urns.extend(self._inflate_partial_urn(partial_urn, account_id, regions))

        for partial_urn in self.delete_urns:
            new_action_set.delete_urns.extend(self._inflate_partial_urn(partial_urn, account_id, regions))

        return new_action_set

    @staticmethod
    def _inflate_partial_urn(partial_urn: PartialUrn, account_id: str, regions: List[str]) -> List[PartialUrn]:
        if partial_urn.region != "ALL":
            return [partial_urn.copy(account_id=account_id)]

        inflated_partial_urns = []
        for region in regions:
            inflated_partial_urns.append(partial_urn.copy(account_id=account_id, region=region))
        return inflated_partial_urns


class ServiceResourceType(NamedTuple):
    """A resource type including a service that it is member of.

    Attributes:
        service_name: The name of the service
        resource_type: The type of resource
    """

    service_name: str
    name: str


class Relationship(NamedTuple):
    """Specifying the relationship between two resources."""

    partial_urn: PartialUrn
    direction: "RelationshipDirection"


@unique
class RelationshipAccountIdSource(Enum):
    """Enum specifying the source of a relationship's Account ID."""

    UNKNOWN = auto()
    SAME_AS_RESOURCE = auto()


@unique
class RelationshipRegionSource(Enum):
    """Enum specifying the source of a relationship's Region."""

    UNKNOWN = auto()
    SAME_AS_RESOURCE = auto()


@unique
class RelationshipDirection(Enum):
    """Enum specifying the direction of a relationship."""

    OUTBOUND = auto()
    INBOUND = auto()



"""Storage Connector for Gremlin databases."""
import logging
from datetime import datetime
from typing import Any, Dict, Iterator, List, Optional, Union

from gremlin_python.driver.driver_remote_connection import DriverRemoteConnection  # type: ignore
from gremlin_python.process.anonymous_traversal import traversal  # type: ignore
from gremlin_python.process.graph_traversal import __  # type: ignore
from gremlin_python.process.traversal import Cardinality, P, T, Traversal  # type: ignore

from cloudwanderer.models import RelationshipDirection

from ..cloud_wanderer_resource import CloudWandererResource
from ..urn import URN, PartialUrn
from .base_connector import ISO_DATE_FORMAT, BaseStorageConnector

logger = logging.getLogger(__name__)


def generate_primary_label(urn: PartialUrn) -> str:
    """Generate a primary vertex label.

    Arguments:
        urn: The URN to generate the vertex label from.
    """
    return urn.cloud_service_resource_label


class GremlinStorageConnector(BaseStorageConnector):
    """Storage Connector for Gremlin databases."""

    _g: Optional[Traversal] = None
    connection: Optional[DriverRemoteConnection] = None

    def __init__(self, endpoint_url: str, supports_multiple_labels=False, test_prefix: str = "", **kwargs) -> None:
        """Create a GremlinStorageConnector.

        Arguments:
            endpoint_url: The url of the gremlin endpoint to connect to (e.g. ``ws://localhost:8182``)
            supports_multiple_labels: Some GraphDBs (Neptune/Neo4J) support multiple labels on a single vertex.
            test_prefix: A prefix that will be prepended to edge and vertex ids to allow scenario separation.
            **kwargs: Any unspecified args will be pased to the ``DriverRemoteConnection`` object.
        """
        self.endpoint_url = endpoint_url
        self.supports_multiple_labels = supports_multiple_labels
        self.test_prefix = test_prefix
        self.connection_args = kwargs

    def init(self) -> None:
        ...

    @property
    def g(self) -> Traversal:
        if not self.connection:
            self.open()
        if not self._g:
            self._g = traversal().withRemote(self.connection)
        return self._g

    def open(self) -> DriverRemoteConnection:
        if not self.connection:
            logger.debug("Opening connection to %s", self.endpoint_url)
            self.connection = DriverRemoteConnection(f"{self.endpoint_url}/gremlin", "g", **self.connection_args)

    def close(self) -> None:
        logger.debug("Closing gremlin connection")
        if self.connection:
            self.connection.close()
        self.connection = None
        self._g = None

    def write_resource(self, resource: CloudWandererResource) -> None:
        """Persist a single resource to storage.

        Arguments:
            resource (CloudWandererResource): The CloudWandererResource to write.
        """
        self._write_resource(resource)
        self._write_dependent_resource_edges(resource)

    def _write_resource(self, resource: CloudWandererResource) -> None:
        primary_label = generate_primary_label(resource.urn)

        traversal = self._write_vertex(vertex_id=self.generate_vertex_id(resource.urn), vertex_labels=[primary_label])
        traversal = (
            traversal.property(Cardinality.single, "_cloud_name", resource.urn.cloud_name)
            .property(Cardinality.single, "_account_id", resource.urn.account_id)
            .property(Cardinality.single, "_region", resource.urn.region)
            .property(Cardinality.single, "_service", resource.urn.service)
            .property(Cardinality.single, "_resource_type", resource.urn.resource_type)
            .property(Cardinality.single, "_discovery_time", resource.discovery_time.isoformat())
            .property(Cardinality.single, "_urn", str(resource.urn))
        )
        for id_part in resource.urn.resource_id_parts:
            traversal.property(Cardinality.set_, "_resource_id_parts", id_part)
        traversal = self._write_properties(
            traversal=traversal, properties=resource.cloudwanderer_metadata.resource_data
        )
        traversal.next()

        self._write_relationships(resource)
        self._clean_up_relationships(urn=resource.urn, cutoff=resource.discovery_time)
        if not resource.urn.is_partial:
            self._repoint_vertex_edges(vertex_label=primary_label, new_resource_urn=resource.urn)

    def _write_dependent_resource_edges(self, resource: CloudWandererResource) -> None:
        for dependent_urn in resource.dependent_resource_urns:
            self._write_edge(
                edge_id=self.generate_edge_id(resource.urn, dependent_urn),
                edge_label="has",
                source_vertex_id=self.generate_vertex_id(resource.urn),
                destination_vertex_id=self.generate_vertex_id(dependent_urn),
                owner_id=self.generate_vertex_id(resource.urn),
                discovery_time=resource.discovery_time,
            ).next()

    def _clean_up_relationships(self, urn: PartialUrn, cutoff: datetime) -> None:
        logger.debug("Cleaning up edges owned by %s discovered before %s", self.generate_vertex_id(urn), cutoff)
        (
            self.g.V(self.generate_vertex_id(urn))
            .bothE()
            .as_("edge")
            .has("_edge_owner", self.generate_vertex_id(urn))
            .where(__.values("_discovery_time").is_(P.lt(cutoff.isoformat())))
            .select("edge")
            .drop()
        ).iterate()

    def _write_relationships(self, resource: CloudWandererResource) -> None:
        for relationship in resource.relationships:
            inferred_partner_urn = relationship.partial_urn
            try:
                pre_existing_resource_id = self._lookup_resource(relationship.partial_urn).next().id
            except StopIteration:
                pre_existing_resource_id = None

            if pre_existing_resource_id:
                logger.debug("writing relationship with pre_existing_resource_id %s", pre_existing_resource_id)
                self._write_relationship_edge(
                    resource_urn=resource.urn,
                    relationship_resource_urn=pre_existing_resource_id,
                    direction=relationship.direction,
                    discovery_time=resource.discovery_time,
                )
                if pre_existing_resource_id != inferred_partner_urn:
                    self._delete_relationship_edge(
                        resource_urn=resource.urn,
                        relationship_resource_urn=inferred_partner_urn,
                        direction=relationship.direction,
                    )
                continue
            logger.debug("Writing inferred resource %s", inferred_partner_urn)
            self._write_resource(CloudWandererResource(urn=relationship.partial_urn, resource_data={}))
            self._write_relationship_edge(
                resource_urn=resource.urn,
                relationship_resource_urn=inferred_partner_urn,
                direction=relationship.direction,
                discovery_time=resource.discovery_time,
            )

    def _repoint_vertex_edges(self, vertex_label: str, new_resource_urn: Union[URN, PartialUrn]) -> None:
        # https://tinkerpop.apache.org/docs/current/recipes/#edge-move

        resources_of_the_same_type = self.g.V().as_("old_vertex").hasLabel(vertex_label)
        for id_part in new_resource_urn.resource_id_parts:
            resources_of_the_same_type.has("_resource_id_parts", id_part)
        resources_with_same_id_but_unknown = (
            resources_of_the_same_type.or_(__.has("_account_id", "unknown"), __.has("_region", "unknown"))
        ).toList()

        for old_vertex in resources_with_same_id_but_unknown:
            # Outbound
            old_vertices_outbound_edges = self.g.V(old_vertex).outE().as_("e1")
            old_outbound_edges_partner_vertex = old_vertices_outbound_edges.inV().as_("b")

            new_vertex = old_outbound_edges_partner_vertex.V(self.generate_vertex_id(new_resource_urn)).as_(
                "new_vertex"
            )
            add_old_outbound_edges_to_new_vertex = (
                new_vertex.addE("has")
                .to("b")
                .as_("e2")
                .sideEffect(
                    __.select("e1")
                    .properties()
                    .unfold()
                    .as_("p")
                    .select("e2")
                    .property(__.select("p").key(), __.select("p").value())
                )
            )
            add_old_outbound_edges_to_new_vertex.select("e1").drop().iterate()
            # Inbound
            old_vertices_inbound_edges = self.g.V(old_vertex).select("old_vertex").inE().as_("old_inbound_edge")
            old_inbound_edges_partner_vertex = old_vertices_inbound_edges.inV().as_("c")

            new_vertex = old_inbound_edges_partner_vertex.select("new_vertex")
            add_old_inbound_edges_to_new_vertex = (
                new_vertex.addE("has")
                .from_("c")
                .as_("new_inbound_edge")
                .sideEffect(
                    __.select("old_inbound_edge")
                    .properties()
                    .unfold()
                    .as_("p")
                    .select("new_inbound_edge")
                    .property(__.select("p").key(), __.select("p").value())
                )
            )
            add_old_inbound_edges_to_new_vertex.select("old_inbound_edge").drop().iterate()

            # Delete old vertex
            self.g.V(old_vertex).drop().iterate()

    def _delete_relationship_edge(
        self, resource_urn: PartialUrn, relationship_resource_urn: PartialUrn, direction: RelationshipDirection
    ) -> None:
        if direction == RelationshipDirection.INBOUND:
            self._delete_edge(self.generate_edge_id(relationship_resource_urn, resource_urn))
        else:
            self._delete_edge(self.generate_edge_id(resource_urn, relationship_resource_urn))

    def _write_relationship_edge(
        self,
        resource_urn: PartialUrn,
        relationship_resource_urn: PartialUrn,
        direction: RelationshipDirection,
        discovery_time: datetime,
    ) -> None:
        if direction == RelationshipDirection.INBOUND:
            self._write_edge(
                edge_id=self.generate_edge_id(relationship_resource_urn, resource_urn),
                edge_label="has",
                source_vertex_id=self.generate_vertex_id(relationship_resource_urn),
                destination_vertex_id=self.generate_vertex_id(resource_urn),
                owner_id=self.generate_vertex_id(resource_urn),
                discovery_time=discovery_time,
            ).next()
        else:
            self._write_edge(
                edge_id=self.generate_edge_id(resource_urn, relationship_resource_urn),
                edge_label="has",
                source_vertex_id=self.generate_vertex_id(resource_urn),
                destination_vertex_id=self.generate_vertex_id(relationship_resource_urn),
                owner_id=self.generate_vertex_id(resource_urn),
                discovery_time=discovery_time,
            ).next()

    def _lookup_resource(self, partial_urn: PartialUrn) -> Traversal:
        vertex_label = generate_primary_label(partial_urn)
        logger.debug("looking up resource with label %s", vertex_label)
        traversal = (
            self.g.V()
            .hasLabel(vertex_label)
            .has("_cloud_name", partial_urn.cloud_name)
            .has("_service", partial_urn.service)
            .has("_resource_type", partial_urn.resource_type)
        )
        for id_part in partial_urn.resource_id_parts:
            traversal.has("_resource_id_parts", id_part)
        if partial_urn.account_id != "unknown":
            traversal.has("_account_id", partial_urn.account_id)
        if partial_urn.region != "unknown":
            traversal.has("_region", partial_urn.region)
        return traversal

    def _write_vertex(self, vertex_id: str, vertex_labels: List[str]) -> Traversal:
        logger.debug("Writing vertex %s", vertex_id)
        if self.supports_multiple_labels:
            vertex_label = "::".join(vertex_labels)
        else:
            vertex_label = vertex_labels[0]
        return self.g.V(vertex_id).fold().coalesce(__.unfold(), __.addV(vertex_label).property(T.id, vertex_id))

    def _write_properties(self, traversal: Traversal, properties: Dict[str, Any]) -> Traversal:
        logger.debug("Writing properties: %s", properties)
        for property_name, property_value in properties.items():
            traversal = traversal.property(Cardinality.single, str(property_name), str(property_value))
        return traversal

    def _write_edge(
        self,
        edge_id: str,
        edge_label: str,
        source_vertex_id: str,
        destination_vertex_id: str,
        owner_id: str,
        discovery_time: datetime,
    ) -> Traversal:
        logger.debug("Writing edge between %s and %s", source_vertex_id, destination_vertex_id)
        return (
            self.g.V(source_vertex_id)
            .as_("source")
            .V(destination_vertex_id)
            .coalesce(
                __.inE().where(__.outV().as_("source")),
                __.addE(edge_label)
                .from_("source")
                .property(T.id, edge_id)
                .property("_edge_owner", owner_id)
                .property("_discovery_time", discovery_time.isoformat()),
            )
        )

    def _delete_edge(self, edge_id: str) -> Traversal:
        logger.debug("Deleting edge %s", edge_id)
        self.g.E(edge_id).drop().iterate()

    def read_all(self) -> Iterator[dict]:
        """Return all records from storage."""

    def read_resource(self, urn: URN) -> Optional[CloudWandererResource]:
        """Return a resource matching the supplied urn from storage.

        Arguments:
            urn (URN): The AWS URN of the resource to return
        """
        return next(self.read_resources(urn=urn))

    def read_resources(
        self,
        cloud_name: str = None,
        account_id: str = None,
        region: str = None,
        service: str = None,
        resource_type: str = None,
        urn: Union[URN, PartialUrn] = None,
    ) -> Iterator["CloudWandererResource"]:
        """Yield a resource matching the supplied urn from storage.

        All arguments are optional.

        Arguments:
            cloud_name: The name of the cloud in question (e.g. ``aws``)
            urn: The AWS URN of the resource to return
            account_id: Cloud Account ID (e.g. ``111111111111``)
            region: AWS region (e.g. ``'eu-west-2'``)
            service: Service name (e.g. ``'ec2'``)
            resource_type: Resource Type (e.g. ``'instance'``)
        """
        if not urn:
            urn = PartialUrn(
                cloud_name=cloud_name or "unknown",
                service=service or "unknown",
                account_id=account_id or "unknown",
                region=region or "unknown",
                resource_type=resource_type or "unknown",
            )
        for vertex in self._lookup_resource(partial_urn=urn).propertyMap().toList():
            yield CloudWandererResource(
                urn=URN.from_string(vertex["_urn"][0].value),
                resource_data=_normalise_gremlin_attrs(vertex),
                discovery_time=datetime.strptime(vertex["_discovery_time"][0].value, ISO_DATE_FORMAT),
            )

    def delete_resource(self, urn: URN) -> None:
        """Delete this resource and all its resource attributes.

        Arguments:
            urn (URN): The URN of the resource to delete
        """
        logger.debug("Deleting resource %s", urn)
        self.g.V(self.generate_vertex_id(urn)).drop().iterate()

    def delete_resource_of_type_in_account_region(
        self,
        cloud_name: str,
        service: str,
        resource_type: str,
        account_id: str,
        region: str,
        cutoff: Optional[datetime],
    ) -> None:
        """Delete resources of type in account and region unless in list of URNs.

        This is used primarily to clean up old resources.

        Arguments:
            cloud_name: The name of the cloud in question (e.g. ``aws``)
            account_id: Cloud Account ID (e.g. ``111111111111``)
            region: Cloud region (e.g. ``'eu-west-2'``)
            service: Service name (e.g. ``'ec2'``)
            resource_type: Resource Type (e.g. ``'instance'``)
            cutoff: Delete any resource discovered before this time
        """
        partial_urn = PartialUrn(
            cloud_name=cloud_name,
            service=service,
            account_id=account_id,
            region=region,
            resource_type=resource_type,
        )
        logger.debug("Deleting resources that match %s that were discovered before %s", partial_urn, cutoff)
        traversal = self._lookup_resource(partial_urn=partial_urn)
        if cutoff:
            traversal.where(__.values("_discovery_time").is_(P.lt(cutoff.isoformat())))
        traversal.drop().iterate()

    def generate_vertex_id(self, urn: PartialUrn) -> str:
        """Generate a vertex id.

        Arguments:
            urn: The URN of the vertex to create.
        """
        return f"{self.test_prefix}{urn}"

    def generate_edge_id(self, source_urn: PartialUrn, destination_urn: PartialUrn) -> str:
        """Generate a primary edge id.

        Arguments:
            source_urn: The URN of the resource we're generating an edge from.
            destination_urn: The URN of the resource we're generating an edge to.
        """
        return f"{self.test_prefix}{source_urn}#{destination_urn}"


def _normalise_gremlin_attrs(raw_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Remove any underscore prefixed keys as these are attributes we use to identify the DynamoDB record.

    Arguments:
        raw_dict: The raw dictionary of the DynamoDB record that needs cleaning.
    """
    # TODO: add list support in gremlin
    normalised_dict = {k: v[0].value for k, v in raw_dict.items() if not k.startswith("_")}
    return normalised_dict



"""A standardised interface for interacting with AWS.

Provides simpler methods for :class:`~.cloud_wanderer.CloudWanderer` to call.
"""

import logging
from typing import TYPE_CHECKING, Iterator, List, Optional, cast

import botocore
from boto3.resources.base import ServiceResource

from ..cloud_wanderer_resource import CloudWandererResource
from ..exceptions import UnsupportedResourceTypeError
from ..models import ActionSet, ServiceResourceType, TemplateActionSet
from ..urn import URN
from .aws_services import AWS_SERVICES
from .session import CloudWandererBoto3Session

if TYPE_CHECKING:

    from .stubs.resource import CloudWandererServiceResource

from .boto3_loaders import ResourceMap

logger = logging.getLogger(__name__)


class CloudWandererAWSInterface:
    """Simplifies lookup of Boto3 services and resources."""

    limit_resources = None

    def __init__(
        self,
        cloudwanderer_boto3_session: Optional[CloudWandererBoto3Session] = None,
    ) -> None:
        """Simplifies lookup of Boto3 services and resources.

        Arguments:
            cloudwanderer_boto3_session:
                A CloudWandererBoto3Session session, if not provided the default will be used.
        """
        self.cloudwanderer_boto3_session = cloudwanderer_boto3_session or CloudWandererBoto3Session()

    def get_enabled_regions(self) -> List[str]:
        """Return the list of regions enabled.

        Fulfils the interface requirements for :class:`cloudwanderer.cloud_wanderer.CloudWanderer` to call.
        """
        return self.cloudwanderer_boto3_session.get_enabled_regions()

    def get_resource(self, urn: URN, include_dependent_resources: bool = True) -> Iterator[CloudWandererResource]:
        """Yield the resource picked out by this URN and optionally its subresources.

        Arguments:
            urn (URN): The urn of the resource to get.
            include_dependent_resources: Whether or not to additionally yield the dependent_resources of the resource.

        Raises:
            UnsupportedResourceTypeError: Occurs when we try to get an unsupported resource type.
            botocore.exceptions.ClientError: Raises from Boto3 client.
        """
        try:
            service = self.cloudwanderer_boto3_session.resource(
                service_name=cast(AWS_SERVICES, urn.service), region_name=urn.region
            )
            if service.service_map.is_global_service and service.service_map.global_service_region != urn.region:
                logger.info(
                    "Creating service in %s instead of the resource's %s region because the service has a global API",
                    service.service_map.global_service_region,
                    urn.region,
                )
                service = self.cloudwanderer_boto3_session.resource(
                    service_name=cast(AWS_SERVICES, urn.service), region_name=service.service_map.global_service_region
                )
            resource = service.resource(resource_type=urn.resource_type, identifiers=urn.resource_id_parts)
            if not hasattr(resource, "load"):
                raise UnsupportedResourceTypeError(f"Resource type {urn.resource_type} doesn't support get_resource()")
            logger.info("Loading resource data.")
            resource.load()

        except botocore.exceptions.ClientError as ex:
            error_code = ex.response.get("ResponseMetadata", {}).get("HTTPStatusCode")
            if not error_code:
                raise
            if error_code == 404:
                return None
            if error_code >= 400 and error_code < 500:
                return None
            raise
        if not resource.meta.data:
            logger.warning("Found no data for %s", urn)
            return None

        dependent_resource_urns = []
        if include_dependent_resources:
            for dependent_resource_type in resource.dependent_resource_types:
                logger.info(
                    "Getting %s %s dependent resources from %s for %s",
                    urn.service,
                    dependent_resource_type,
                    urn.region,
                    resource.get_urn().resource_id,
                )
                for dependent_resource in resource.collection(resource_type=dependent_resource_type):
                    logger.debug("Found %s", dependent_resource)
                    if not dependent_resource.meta.data and hasattr(dependent_resource, "load"):
                        dependent_resource.load()
                    urn = dependent_resource.get_urn()
                    dependent_resource_urns.append(urn)
                    yield CloudWandererResource(
                        urn=urn,
                        resource_data={
                            **dependent_resource.normalized_raw_data,
                            **dependent_resource.get_secondary_attributes_map(),
                        },
                        parent_urn=resource.get_urn(),
                        relationships=dependent_resource.relationships,
                    )
        yield CloudWandererResource(
            urn=resource.get_urn(),
            resource_data={**resource.normalized_raw_data, **resource.get_secondary_attributes_map()},
            dependent_resource_urns=dependent_resource_urns,
            relationships=resource.relationships,
        )

    def get_resources(
        self,
        service_name: str,
        resource_type: str,
        region: str,
        **kwargs,
    ) -> Iterator[CloudWandererResource]:
        """Return all resources of resource_type from Boto3.

        Arguments:
            service_name (str): The name of the service to get resource for (e.g. ``'ec2'``)
            resource_type (str): The type of resource to get resources of (e.g. ``'instance'``)
            region (str): The region to get resources of (e.g. ``'eu-west-1'``)
            **kwargs: Additional keyword arguments will be passed down to the Boto3 client.

        Raises:
            botocore.exceptions.ClientError: Occurs if the Boto3 Client Errors.
        """
        service_name = cast(AWS_SERVICES, service_name)
        logger.info("Getting %s %s resources from %s", service_name, resource_type, region)
        service = self.cloudwanderer_boto3_session.resource(service_name=service_name, region_name=region)
        resource_map: ResourceMap = service.service_map.get_resource_map(resource_type)

        try:
            for resource in service.collection(resource_type=resource_type, filters=resource_map.default_filters):
                dependent_resource_urns = []
                for dependent_resource_type in resource.dependent_resource_types:
                    logger.info(
                        "Getting %s %s dependent resources from %s for %s",
                        service_name,
                        dependent_resource_type,
                        region,
                        resource.get_urn().resource_id,
                    )
                    for dependent_resource in resource.collection(resource_type=dependent_resource_type):
                        logger.debug("Found %s", dependent_resource)
                        if not dependent_resource.meta.data and hasattr(dependent_resource, "load"):
                            dependent_resource.load()
                        urn = dependent_resource.get_urn()
                        dependent_resource_urns.append(urn)
                        yield CloudWandererResource(
                            urn=urn,
                            resource_data={
                                **dependent_resource.normalized_raw_data,
                                **dependent_resource.get_secondary_attributes_map(),
                            },
                            parent_urn=resource.get_urn(),
                            relationships=dependent_resource.relationships,
                        )
                yield CloudWandererResource(
                    urn=resource.get_urn(),
                    resource_data={**resource.normalized_raw_data, **resource.get_secondary_attributes_map()},
                    dependent_resource_urns=dependent_resource_urns,
                    relationships=resource.relationships,
                )
        except botocore.exceptions.EndpointConnectionError:
            logger.info("%s %s not supported in %s", service_name, resource_type, region)
            return
        except botocore.exceptions.ClientError as ex:
            if ex.response["Error"]["Code"] == "InvalidAction":
                logger.info("%s %s not supported in %s", service_name, resource_type, region)
                return
            raise

    def get_resource_discovery_actions(
        self, regions: List[str] = None, service_resource_types: List[ServiceResourceType] = None
    ) -> List[ActionSet]:
        service_resource_types = service_resource_types or []
        discovery_regions = regions or self.cloudwanderer_boto3_session.get_enabled_regions()

        service_names = [resource_type.service_name for resource_type in service_resource_types]
        action_sets = []
        service_names = service_names or self.cloudwanderer_boto3_session.get_available_resources()
        logger.debug("Getting actions for: %s", service_names)
        for service_name in service_names:
            service_specific_resource_types = [
                resource_type.name
                for resource_type in service_resource_types
                if resource_type.service_name == service_name
            ]
            service = self.cloudwanderer_boto3_session.resource(service_name=cast(AWS_SERVICES, service_name))
            action_sets.extend(
                self._get_discovery_action_templates_for_service(
                    service=service, resource_types=service_specific_resource_types, discovery_regions=discovery_regions
                )
            )

        return self._inflate_action_set_regions(action_sets)

    def get_all_empty_resources(self, include_dependent_resource=False) -> Iterator[ServiceResource]:
        """Return an ``empty_resource=True`` ServiceResource object for each resource type.

        Arguments:
            include_dependent_resource: Whether or not dependent resources should be returned.
        """
        for service_name in self.cloudwanderer_boto3_session.get_available_resources():
            service = self.cloudwanderer_boto3_session.resource(service_name)  # type: ignore
            for resource_type in service.resource_types:
                resource = service.resource(resource_type, empty_resource=True)
                yield resource
                if not include_dependent_resource:
                    continue
                for dependent_resource_type in resource.dependent_resource_types:
                    yield service.resource(dependent_resource_type, empty_resource=True)

    def _inflate_action_set_regions(self, action_set_templates: List[TemplateActionSet]) -> List[ActionSet]:
        enabled_regions = self.cloudwanderer_boto3_session.get_enabled_regions()
        result = []
        for action_set_template in action_set_templates:
            result.append(
                action_set_template.inflate(
                    regions=enabled_regions, account_id=self.cloudwanderer_boto3_session.get_account_id()
                )
            )

        return result

    def _get_discovery_action_templates_for_service(
        self, service: "CloudWandererServiceResource", resource_types: List[str], discovery_regions: List[str]
    ) -> List[TemplateActionSet]:
        logger.debug("Getting resource_types for %s", service.service_name)

        if resource_types:
            service_resource_types = list(set(resource_types) & set(service.resource_types))
        else:
            service_resource_types = service.resource_types
        action_templates = []
        for resource_type in service_resource_types:
            resource = service.resource(resource_type, empty_resource=True)
            action_templates.extend(
                self._get_discovery_action_templates_for_resource(
                    service=service, resource=resource, discovery_regions=discovery_regions
                )
            )
        return action_templates

    def _get_discovery_action_templates_for_resource(
        self,
        service: "CloudWandererServiceResource",
        resource: "CloudWandererServiceResource",
        discovery_regions: List[str],
    ) -> List[TemplateActionSet]:
        action_templates = []

        logger.debug("Getting actions for %s in %s", resource.resource_type, discovery_regions)

        action_templates = resource.get_discovery_action_templates(discovery_regions=discovery_regions)
        if not action_templates:
            return []
        logger.debug("getting actions for: %s", resource.dependent_resource_types)
        for dependent_resource_type in resource.dependent_resource_types:
            dependent_resource = service.resource(dependent_resource_type, empty_resource=True)
            action_templates.extend(
                dependent_resource.get_discovery_action_templates(discovery_regions=discovery_regions)
            )

        return action_templates



"""Boto3 Loaders.

Loaders and data classes to load custom CloudWanderer Boto3 service definitions and merge them with the boto3
provided ones. This allows cloudwanderer to extend Boto3 to support AWS resources that it does not support natively.
We can do this quite easily because CloudWanderer only needs a fraction of the functionality that native
Boto3 resources provide (i.e. the description of the resources).
"""

import json
import logging
import os
import pathlib
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional

import boto3
import botocore
import jmespath  # type: ignore
from boto3.resources.base import ServiceResource
from botocore.exceptions import DataNotFoundError, UnknownServiceError  # type: ignore
from botocore.loaders import Loader

from ..cache_helpers import memoized_method
from ..exceptions import MalformedFileError, UnsupportedServiceError
from ..models import RelationshipAccountIdSource, RelationshipDirection, RelationshipRegionSource
from ..utils import camel_to_snake, snake_to_pascal

logger = logging.getLogger(__name__)


class CustomServiceLoader:
    """A class to load custom services."""

    def __init__(self, definition_path: str = "resource_definitions") -> None:
        self.service_definitions_path = os.path.join(pathlib.Path(__file__).parent.absolute(), definition_path)

    @memoized_method()
    def get_service_definition(self, service_name: str, type_name: str, api_version: str) -> dict:
        path = Path(service_name) / Path(api_version) / Path(f"{type_name}.json")
        full_path = self.service_definitions_path / path
        logger.debug("CustomServiceLoader get_service_definition loaded path %s", full_path)
        try:
            with open(full_path, "r") as file:
                return json.load(file)
        except FileNotFoundError:
            raise UnsupportedServiceError(f"{service_name} does not exist as a custom CloudWanderer service.")
        except json.decoder.JSONDecodeError as ex:
            raise MalformedFileError(f"{service_name} has an invalid file at {full_path}.") from ex

    def list_api_versions(self, service_name: str, type_name: str) -> List[str]:
        path = self.service_definitions_path / Path(service_name)
        try:
            api_versions = [
                d
                for d in os.listdir(path)
                if os.path.isdir(os.path.join(path, d)) and os.path.isfile(os.path.join(path, d, f"{type_name}.json"))
            ]
        except FileNotFoundError:
            raise UnsupportedServiceError(f"{service_name} does not exist as a custom CloudWanderer service.")

        return sorted(api_versions)

    @property
    def available_services(self) -> List[str]:
        """Return a list of available snake_case service names."""
        possible_services = [
            d
            for d in os.listdir(self.service_definitions_path)
            if os.path.isdir(os.path.join(self.service_definitions_path, d))
        ]
        return sorted(possible_services)


class MergedServiceLoader(Loader):
    """A class to merge the services from a custom service loader with those of Boto3."""

    def __init__(self) -> None:
        self.custom_service_loader = CustomServiceLoader()
        botocore_session = botocore.session.get_session()
        self.botocore_loader = botocore_session.get_component("data_loader")
        boto3_data_path = os.path.join(os.path.dirname(boto3.__file__), "data")
        self.botocore_loader.search_paths.append(boto3_data_path)

    def list_available_services(self, type_name: str = "resources-1") -> List[str]:
        _ = type_name
        """Return a list of service names that can be loaded."""
        return list(set(self.cloudwanderer_available_services + self.boto3_available_services))

    def determine_latest_version(self, service_name: str, type_name: str) -> str:
        return max(self.list_api_versions(service_name, type_name))

    @property
    def cloudwanderer_available_services(self) -> List[str]:
        """Return a list of services defined by CloudWanderer."""
        return self.custom_service_loader.available_services

    @property
    def boto3_available_services(self) -> List[str]:
        """Return a list of services defined by Boto3."""
        return self.botocore_loader.list_available_services(type_name="resources-1")

    def list_api_versions(self, service_name: str, type_name: str) -> List[str]:
        logger.debug("list_api_version, service_name: %s, type_name: %s", service_name, type_name)
        try:
            boto3_api_versions = self.botocore_loader.list_api_versions(service_name=service_name, type_name=type_name)
        except DataNotFoundError:
            boto3_api_versions = []
        try:
            custom_api_versions = self.custom_service_loader.list_api_versions(
                service_name=service_name, type_name=type_name
            )
        except UnsupportedServiceError:
            custom_api_versions = []

        if not boto3_api_versions and not custom_api_versions:
            raise UnsupportedServiceError(
                f"{service_name} is not supported by either Boto3 or CloudWanderer's "
                f"custom services for {type_name}.json"
            )

        if custom_api_versions:
            return custom_api_versions

        return boto3_api_versions

    @memoized_method()
    def load_service_model(self, service_name: str, type_name: str, api_version: Optional[str] = None) -> OrderedDict:
        logger.debug(
            "botocore_loaders load_service_model service_name %s, type_name %s, api_version %s",
            service_name,
            type_name,
            api_version,
        )
        if not api_version:
            api_version = self.determine_latest_version(service_name=service_name, type_name=type_name)
        try:
            boto3_definition = self.botocore_loader.load_service_model(
                service_name, type_name=type_name, api_version=api_version
            )
        except UnknownServiceError:
            boto3_definition = OrderedDict()

        custom_service_definition = self._get_custom_service_definition(
            service_name, type_name=type_name, api_version=api_version
        )

        if not boto3_definition and not custom_service_definition:
            raise UnsupportedServiceError(
                f"{service_name} is not supported by either Boto3 or CloudWanderer's custom "
                f"services for api_version {api_version}, type {type_name}"
            )
        merged_dict = OrderedDict(
            {
                "service": OrderedDict(
                    {
                        **custom_service_definition.get("service", {}),
                        **{
                            "hasMany": OrderedDict(
                                {
                                    **(jmespath.search("service.hasMany", boto3_definition) or OrderedDict({})),
                                    **(
                                        jmespath.search("service.hasMany", custom_service_definition) or OrderedDict({})
                                    ),
                                }
                            )
                        },
                    }
                ),
                "resources": OrderedDict(
                    {
                        **boto3_definition.get("resources", OrderedDict({})),
                        **custom_service_definition.get("resources", OrderedDict({})),
                    }
                ),
            }
        )
        # logger.warning(merged_dict['resources'].get('RouteTable'))
        return merged_dict

    def _get_custom_service_definition(self, service_name: str, type_name: str, api_version: str) -> dict:
        """Get the custom CloudWanderer definition for service_name so we can merge it with Boto3's.

        Arguments:
            service_name: The PascalCase name of the service to get the definition of.
            api_version: The version of the api to get the definition of.
            type_name: The type of definition to get.
        """
        if not api_version:
            api_version = self.determine_latest_version(service_name=service_name, type_name=type_name)
        try:
            return self.custom_service_loader.get_service_definition(
                service_name=service_name, type_name=type_name, api_version=api_version
            )
        except UnsupportedServiceError:
            return {}


# # TODO: Normalising servicemapping as servicemap


class ServiceMap(NamedTuple):
    """Specification for additional CloudWanderer specific metadata about a Boto3 service."""

    name: str
    resource_definition: dict
    global_service: bool
    global_service_region: str
    service_definition: dict

    @property
    def is_global_service(self) -> bool:
        """Return whether this service is a global service."""
        return self.global_service

    @property
    def is_default_service(self) -> bool:
        """Return True if this service has no definition and should return default values."""
        return self.service_definition == {}

    def get_resource_map(self, resource_type: str) -> "ResourceMap":
        """Return the resource map given a snake_case resource name.

        Arguments:
            resource_type: The snake_case name of the resource map to get.
        """
        boto3_resource_name = snake_to_pascal(resource_type)
        return ResourceMap.factory(
            service_map=self,
            definition=self.resource_definition.get(boto3_resource_name, {}),
        )

    @classmethod
    def factory(cls, name: str, definition: dict) -> "ServiceMap":
        service_definition = definition.get("service", {})
        return cls(
            name=name,
            global_service=service_definition.get("globalService", False),
            global_service_region=service_definition.get("globalServiceRegion"),
            service_definition=service_definition,
            resource_definition=definition.get("resources", {}),
        )


class ResourceMap(NamedTuple):
    """Specification for additional CloudWanderer specific metadata about a Boto3 resource type.

    Attributes:
        region_request:
            An optional definition for how to perform a secondary query to
            discover the region in which this resource exists.
        parent_resource_type:
            The snake_case resource type of the parent (if this is a subresource).
        ignored_subresources:
            A list of snake case subresources which exist in the Boto3 definition but should be ignored.
        default_filters:
            A dict of arguments to supply to the API Method used when enumerating this resource type.
            This can be overridden by the user with the ``filters`` argument.
        service_map:
            A link back to the parent :class:`ServiceMap` object.
        requires_load_for_full_metadata:
            If set ``resource.load()`` is called when fetching a list of these resources.
            Resources in a big increase in the number of API calls as one must be made for *each* resource.
        regional_resource:
            Whether or not this resource exists in every region.
    """

    type: Optional[str]
    region_request: Optional["ResourceRegionRequest"]
    ignored_subresources: list
    default_filters: Dict[str, Any]
    service_map: ServiceMap
    relationships: List["RelationshipSpecification"]
    secondary_attribute_maps: List["SecondaryAttributeMap"]
    requires_load_for_full_metadata: bool = False
    regional_resource: bool = True

    @classmethod
    def factory(
        cls,
        service_map: ServiceMap,
        definition: Dict[str, Any],
    ) -> "ResourceMap":
        return cls(
            type=definition.get("type"),
            region_request=ResourceRegionRequest.factory(definition.get("regionRequest")),
            ignored_subresources=definition.get("ignoredSubresources", []),
            requires_load_for_full_metadata=definition.get("requiresLoadForFullMetadata", False),
            regional_resource=definition.get("regionalResource", True),
            default_filters=definition.get("defaultFilters", {}),
            service_map=service_map,
            relationships=[
                RelationshipSpecification.factory(relationship_specification)
                for relationship_specification in definition.get("relationships", [])
            ],
            secondary_attribute_maps=[
                SecondaryAttributeMap(source_path=mapping["sourcePath"], destination_name=mapping["destinationName"])
                for mapping in definition.get("secondaryAttributeMaps", [])
            ],
        )

    def should_query_resources_in_region(self, region: str) -> bool:
        """Return whether this resource should be queried from this region.

        Arguments:
            region: The region in which to query the resource.
        """
        if not self.service_map.is_global_service:
            return True
        return self.service_map.is_global_service and self.service_map.global_service_region == region


class ResourceRegionRequest(NamedTuple):
    """Specification for a request to get a resource's region."""

    operation: str
    params: list
    path_to_region: str
    default_value: str

    @classmethod
    def factory(cls, definition: Optional[Dict[str, Any]]) -> Optional["ResourceRegionRequest"]:
        if not definition:
            return None
        return cls(
            operation=definition["operation"],
            params=[ResourceRegionRequestParam.factory(param) for param in definition["params"]],
            path_to_region=definition["pathToRegion"],
            default_value=definition["defaultValue"],
        )

    def build_params(self, resource: ServiceResource) -> dict:
        """Return the params required to query the resource's region from the Boto3 resource's attributes.

        Arguments:
            resource: The Boto3 Resource to build params for.
        """
        params = {}
        for param in self.params:
            params[param.target] = self._get_param_value(resource, param)
        return params

    def _get_param_value(self, resource: ServiceResource, param: "ResourceRegionRequestParam") -> Any:
        if param.source == "resourceAttribute":
            return getattr(resource, param.name)
        raise AttributeError(f"Invalid param source {param.source}")


class ResourceRegionRequestParam(NamedTuple):
    """Specification for a request param data model."""

    target: str
    source: str
    name: str

    @classmethod
    def factory(cls, definition: dict) -> "ResourceRegionRequestParam":
        return cls(target=definition["target"], source=definition["source"], name=definition["name"])


class RelationshipSpecification(NamedTuple):
    """Specification for a relationship between two resources."""

    base_path: str
    id_parts: List["IdPartSpecification"]
    service: str
    resource_type: str
    region_source: RelationshipRegionSource
    account_id_source: RelationshipAccountIdSource
    direction: RelationshipDirection

    @classmethod
    def factory(cls, definition) -> "RelationshipSpecification":
        return cls(
            base_path=definition["basePath"],
            direction=RelationshipDirection[camel_to_snake(definition["direction"])],
            id_parts=[IdPartSpecification.factory(id_part) for id_part in definition["idParts"]],
            service=definition["service"],
            resource_type=definition["resourceType"],
            region_source=RelationshipRegionSource[camel_to_snake(definition["regionSource"])],
            account_id_source=RelationshipAccountIdSource[camel_to_snake(definition["accountIdSource"])],
        )


class IdPartSpecification(NamedTuple):
    """Specification for getting the ID parts of a resource's relationship with another resource."""

    path: str
    regex_pattern: str

    @classmethod
    def factory(cls, definition) -> "IdPartSpecification":
        return cls(path=definition["path"], regex_pattern=definition.get("regexPattern", ""))


class SecondaryAttributeMap(NamedTuple):
    """Specification for mapping attributes contained in a secondary attribute to its parent's resource."""

    source_path: str
    destination_name: str



"""Storage Connectors for CloudWanderer."""
from .base_connector import BaseStorageConnector
from .dynamodb import DynamoDbConnector
from .gremlin import GremlinStorageConnector
from .memory import MemoryStorageConnector

__all__ = ["DynamoDbConnector", "MemoryStorageConnector", "BaseStorageConnector", "GremlinStorageConnector"]



"""The CloudWanderer AWS Interface."""
from .interface import CloudWandererAWSInterface
from .session import CloudWandererBoto3Session

__all__ = ["CloudWandererAWSInterface", "CloudWandererBoto3Session"]



"""Autogenerated Literals."""
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

# TODO: Autogenerate
AWS_SERVICES = Literal["cloudwatch", "sns"]



#!/usr/bin/env python
"""Setup CloudWanderer package."""
import re
from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = re.sub(r"..\s+doctest\s+::", ".. code-block ::", f.read())

setup(
    version="0.19.5",
    python_requires=">=3.6.0",
    name="cloudwanderer",
    packages=find_packages(include=["cloudwanderer", "cloudwanderer.*"]),
    description="A Python package which wanders across your AWS account and records resources in multiple databases",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Sam Martin",
    author_email="samjackmartin+cloudwanderer@gmail.com",
    url="https://github.com/CloudWanderer-io/CloudWanderer",
    install_requires=["boto3", "jmespath", 'typing_extensions; python_version < "3.8.0"', "gremlinpython"],
    package_data={
        "": ["**/**/**/**/*.json", "py.typed"],
    },
)



"""Collection of loose utility functions."""
import json
import logging
from datetime import datetime
from decimal import Decimal
from typing import Any, Callable, Dict, Optional

from botocore import xform_name

logger = logging.getLogger(__name__)


def exception_logging_wrapper(method: Callable, return_value: object = None, **kwargs) -> Optional[Any]:
    """Log exceptions raised by method_name.

    Increases visibility of multi-threaded exceptions.

    Arguments:
        method (Callable): A method to wrap, for which will we handle and log any exceptions.
        return_value (object): A value to return instead of the method's result.
        **kwargs: Arguments to pass to the method
    """
    try:
        result = method(**kwargs)
    except Exception as ex:
        logger.exception(ex)
        return return_value
    if return_value is None:
        return result
    return return_value


def json_object_hook(dct: dict) -> dict:
    """Clean out empty strings to avoid DynamoDB ValidationExceptions.

    Arguments:
        dct (dict): The dictionary to clean.
    """
    for key, value in dct.items():
        if value == "":
            dct[key] = None
    return dct


def json_default(item: object) -> Optional[object]:
    """JSON object type converter that handles datetime objects.

    Arguments:
        item: The object JSON is trying to serialise.
    """
    if isinstance(item, datetime):
        return item.isoformat()
    return None


def standardise_data_types(resource: dict) -> Dict[str, Any]:
    """Return a dictionary normalised to datatypes acceptable for DynamoDB.

    Arguments:
        resource (dict): The dictionary we're normalising to DynamoDB acceptable data types.
    """
    result = json.loads(json.dumps(resource, default=json_default), object_hook=json_object_hook, parse_float=Decimal)
    return result


def snake_to_pascal(snake_case: str) -> str:
    """Return a PascalCase version of a snake_case name.

    Arguments:
        snake_case: the string to turn into PascalCase.
    """
    snake_case = snake_case.lower()
    return snake_case.replace("_", " ").title().replace(" ", "")


def camel_to_snake(camel, upper=True) -> str:
    """Convert camelCase to snake_case (uppercase by default).

    Arguments:
        camel: Input camelCase string
        upper: Whether to conver to uppercase snake_case
    """
    if upper:
        return xform_name(camel).upper()
    return xform_name(camel)



"""CloudWanderer's Exceptions."""


class BadUrnRegionError(Exception):
    """Raised when an URN is passed to get_resource with a region that is not possible."""


class BadUrnIdentifiersError(Exception):
    """Raised when an URN has the wrong number or type of identifier parts."""


class BadUrnAccountIdError(Exception):
    """Raised when an URN is passed to get_resource with an account other than the one the session is for."""


class BadUrnSubResourceError(Exception):
    """Raised when an URN for a subresource is passed to get resource."""


class GlobalServiceResourceMappingNotFoundError(Exception):
    """Global Service Resource Mapping not Found."""


class ResourceActionDoesNotExistError(Exception):
    """Resource does not exist on this service as supported by CloudWanderer."""


class ResourceNotFoundError(Exception):
    """Requested resource was not found."""


class BadRequestError(Exception):
    """The Cloud API returned a Bad Request Error.

    For some services for some clouds this is the same as a ResourceNotFoundError error.
    """


class UnsupportedResourceTypeError(Exception):
    """The resource type in question is not supported for this operation."""


class UnsupportedServiceError(Exception):
    """The service in question is not supported for this operation."""


class BadServiceMapError(Exception):
    """There was an insconsistency in the service map for this operation."""


class MalformedFileError(Exception):
    """There was a malformed file found."""
